#include <slim_blockchain_handler.hpp>
#include <toolbox.hpp>
#include <message.hpp>
#include <enums.hpp>
#include <algorithm>


void SlimBlockchainHandler::update(Toolbox& tb) {
	//can't be dragging if mstate changes
	if (previous_mstate_ != tb.m_state) {
		clear_blocks_();
	}
	//update tracked mstate
	previous_mstate_ = tb.m_state;

	bool skip_events = false;

	/*Handle Messages*/
	for (Message& m : tb.mailbox) {
		if (m.target == MESSAGETARGET::BLOCKCHAIN && !skip_events) {
			BLOCKCHAINMESSAGE message = std::get<BLOCKCHAINMESSAGE>(m.message);
			switch (message) {
			case BLOCKCHAINMESSAGE::CLEAR:
				clear_blocks_();
				m.handled = true;
				skip_events = true;
				break;
			case BLOCKCHAINMESSAGE::PATTERN_SINGLESLIT:
				clear_blocks_();
				queue_singleslit_(tb);
				m.handled = true;
				skip_events = true;
				break;
			case BLOCKCHAINMESSAGE::PATTERN_DOUBLESLIT:
				clear_blocks_();
				queue_doubleslit_(tb);
				m.handled = true;
				skip_events = true;
				break;
			case BLOCKCHAINMESSAGE::PATTERN_LATTICE:
				clear_blocks_();
				queue_lattice_(tb);
				m.handled = true;
				skip_events = true;
				break;
			case BLOCKCHAINMESSAGE::PATTERN_WAVEGUIDE:
				clear_blocks_();
				queue_waveguide_(tb);
				m.handled = true;
				skip_events = true;
				break;
			case BLOCKCHAINMESSAGE::PATTERN_SSH:
				clear_blocks_();
				queue_ssh_(tb);
				m.handled = true;
				skip_events = true;
				break;
			case BLOCKCHAINMESSAGE::PATTERN_FRESNEL:
				clear_blocks_();
				queue_fresnel_(tb);
				m.handled = true;
				skip_events = true;
				break;
			}
		}
	}

	if (tb.m_state == static_cast<int>(MSTATE::IMMEDIATE) && !skip_events) {
		/*Handle Events*/
		for (Pevent& pev : tb.events) {
			/*Handle event differently depending on where it is*/
			/*Event within wave window*/
			if (in_wave_window_(tb, pev)) {
				switch (pev.type) {
					/*Finger down, place Block, add to dragpairs*/
				case PEVENTTYPE::DOWN:
					blockchain_.push_back(EfficientBlock(tb.block_width, tb.block_height, pev.itcoord_x - tb.block_width / 2, tb.texture_h - (pev.itcoord_y + tb.block_height / 2)));
					dragpairs_.push_back({ &blockchain_.back(), pev.finger_id });
					drawlist_dynamic_.push_back(blockchain_.back().xywh());
					break;
					/*Finger up, find and remove the associated Block and dragpair*/
				case PEVENTTYPE::UP: {
					auto released = std::find_if(dragpairs_.begin(), dragpairs_.end(), [=](auto dragpair) { return dragpair.second == pev.finger_id; });
					if (released != dragpairs_.end()) {
						released->first->request_removal();
						eraselist_dynamic_.push_back(released->first->xywh());
					}
					dragpairs_.remove_if([=](auto dragpair) { return dragpair.second == pev.finger_id; });
					break;
				}
													 /*Finger moved, find and translate associated Block*/
				case PEVENTTYPE::MOVE: {
					auto moved = std::find_if(dragpairs_.begin(), dragpairs_.end(), [=](auto dragpair) { return dragpair.second == pev.finger_id; });
					if (moved != dragpairs_.end()) {
						eraselist_dynamic_.push_back(moved->first->xywh());
						moved->first->translate(pev.itcoord_x - moved->first->width() / 2, tb.texture_h - (pev.itcoord_y + moved->first->height() / 2));
						drawlist_dynamic_.push_back(moved->first->xywh());
					}
					break;
				}
				}
			}
			/*Event outside wave window*/
			else {
				switch (pev.type) {
					/*Finger down outside wave area, don't care*/
				case PEVENTTYPE::DOWN:
					break;
					/*Finger up or moved outside wave area, if associated Block exists, clear it*/
				case PEVENTTYPE::UP:
				case PEVENTTYPE::MOVE: {
					auto released = std::find_if(dragpairs_.begin(), dragpairs_.end(), [=](auto dragpair) { return dragpair.second == pev.finger_id; });
					if (released != dragpairs_.end()) {
						released->first->request_removal();
						eraselist_dynamic_.push_back(released->first->xywh());
					}
					dragpairs_.remove_if([=](auto dragpair) { return dragpair.second == pev.finger_id; });
					break;
				}
				}
			}
		}
	}
	/*Update Blocks*/
	/*TODO: Can reupload blocks depending on need, if necessary*/
	update_blocks_(tb, false);

	/*Toolbox developer region*/
	tb.num_blocks = num_blocks();
}

size_t SlimBlockchainHandler::num_blocks() {
	return blockchain_.size();
}

void SlimBlockchainHandler::clear_blocks_() {
	for (EfficientBlock& b : blockchain_) {
		eraselist_dynamic_.push_back(b.xywh());
		b.request_removal();
	}
	dragpairs_.clear();
	blockchain_.clear();
}

void SlimBlockchainHandler::update_blocks_(const Toolbox& tb, bool reload_all) {
	find_dynamic_duplicates_();
	if (drawlist_dynamic_.size() != 0 || drawlist_static_.size() != 0 || eraselist_dynamic_.size() != 0 || queued_texture_ != 0) {
		glBindVertexArray(vao_);
		glBindBuffer(GL_ARRAY_BUFFER, vbo_);
		glViewport(0, 0, tb.texture_w, tb.texture_h);
		glActiveTexture(GL_TEXTURE0);
		shader_drawblocks_.use();
	}
	std::vector<float> vertices;
	/*Erase dynamic blocks*/
	if (eraselist_dynamic_.size() != 0) {
		glBindFramebuffer(GL_FRAMEBUFFER, fbo_dynamic_);
		glBindTexture(GL_TEXTURE_2D, tb.tex_damp_clean);
		xywhs_to_vertices_(tb, eraselist_dynamic_duplicates_, eraselist_dynamic_, vertices);
		if (vertices.size() != 0) {
			glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(float), vertices.data(), GL_DYNAMIC_DRAW);
			/*Draw on damping*/
			shader_drawblocks_.setFloat("color_multiplier", 1.f);
			for (size_t i = 0; i < eraselist_dynamic_.size(); ++i) {
				glDrawArrays(GL_TRIANGLES, 6 * i, 6);
			}
		}
		eraselist_dynamic_.clear();
	}
	/*Draw static blocks*/
	if (drawlist_static_.size() != 0) {
		glBindFramebuffer(GL_FRAMEBUFFER, fbo_static_);
		glBindTexture(GL_TEXTURE_2D, tb.tex_wave_clean);
		xywhs_to_vertices_(tb, std::vector<bool>(drawlist_static_.size(), false), drawlist_static_, vertices);
		if (vertices.size() != 0) {
			glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(float), vertices.data(), GL_DYNAMIC_DRAW);
			/*Draw on damping*/
			shader_drawblocks_.setFloat("color_multiplier", 0.f);
			for (size_t i = 0; i < drawlist_static_.size(); ++i) {
				glDrawArrays(GL_TRIANGLES, 6 * i, 6);
			}
			/*Draw on wave*/
			shader_drawblocks_.setFloat("color_multiplier", 1.f);
			glBindFramebuffer(GL_FRAMEBUFFER, fbo_wave_);
			for (size_t i = 0; i < drawlist_static_.size(); ++i) {
				glDrawArrays(GL_TRIANGLES, 6 * i, 6);
			}
		}
		drawlist_static_.clear();
	}
	/*Draw dynamic blocks*/
	if (drawlist_dynamic_.size() != 0) {
		glBindFramebuffer(GL_FRAMEBUFFER, fbo_dynamic_);
		glBindTexture(GL_TEXTURE_2D, tb.tex_wave_clean);
		xywhs_to_vertices_(tb, drawlist_dynamic_duplicates_, drawlist_dynamic_, vertices);
		if (vertices.size() != 0) {
			glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(float), vertices.data(), GL_DYNAMIC_DRAW);
			/*Draw on damping*/
			shader_drawblocks_.setFloat("color_multiplier", 0.f);
			for (size_t i = 0; i < drawlist_dynamic_.size(); ++i) {
				glDrawArrays(GL_TRIANGLES, 6 * i, 6);
			}
			/*Draw on wave*/
			shader_drawblocks_.setFloat("color_multiplier", 1.f);
			glBindFramebuffer(GL_FRAMEBUFFER, fbo_wave_);
			for (size_t i = 0; i < drawlist_dynamic_.size(); ++i) {
				glDrawArrays(GL_TRIANGLES, 6 * i, 6);
			}
		}
		drawlist_dynamic_.clear();
	}
	/*Draw textured structures*/
	if (queued_texture_ != 0) {
		glBindFramebuffer(GL_FRAMEBUFFER, fbo_static_);
		glBindTexture(GL_TEXTURE_2D, queued_texture_);
		xywhs_to_vertices_(tb, { false }, { {0, 0, tb.texture_w, tb.texture_h} }, vertices);
		if (vertices.size() != 0) {
			glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(float), vertices.data(), GL_DYNAMIC_DRAW);
			shader_drawblocks_.setFloat("color_multiplier", 1.f);
			glDrawArrays(GL_TRIANGLES, 0, 6);
		}
		queued_texture_ = 0;
	}
	/*Drawing finished, reset OpenGL state*/
	glBindVertexArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);


	/*Clean up the blockchain*/
	/* This doesn't work, because it invalidates all iterators/pointers, even in a list. See https://godbolt.org/z/GzxTExc44
	blockchain_.erase(std::remove_if(blockchain_.begin(), blockchain_.end(), [](const Block& b) {return b.needs_removal(); }), blockchain_.end());
	The following preserves iterator/pointer validity, which is what we need
	*/
	dragpairs_.remove_if([=](const std::pair<EfficientBlock*, SDL_FingerID>& dp) { return dp.first->needs_removal(); });
	blockchain_.remove_if([=](const EfficientBlock& b) { return b.needs_removal(); });
}

bool SlimBlockchainHandler::in_wave_window_(const Toolbox& tb, const Pevent& pev) const {
	return pev.fscoord_x < 1.f - tb.gui_pos;
}

/*Structures*/
/*Note that they don't go into the blockchain, they only exist in the static damping texture*/
/*The correct wave texture for Blocks is tb.tex_wave_1*/
void SlimBlockchainHandler::queue_singleslit_(Toolbox& tb) {
	int block_width = 30;
	int slit_height = 40;
	int block_height = tb.texture_h / 2 - slit_height/2 - 1;
	EfficientBlock b_top(block_width, block_height, tb.texture_w / 3, tb.texture_h / 2 + slit_height / 2);
	EfficientBlock b_bottom(block_width, block_height, tb.texture_w / 3, 0);
	drawlist_static_.push_back(b_top.xywh());
	drawlist_static_.push_back(b_bottom.xywh());
}

void SlimBlockchainHandler::queue_doubleslit_(Toolbox& tb) {
	int block_width = 30;
	int slit_height = 40;
	int mid_block_height = 30;
	int block_height = tb.texture_h / 2 - slit_height - mid_block_height / 2 - 1;
	EfficientBlock b_top(block_width, block_height, tb.texture_w / 3, tb.texture_h / 2 + mid_block_height / 2 + slit_height);
	EfficientBlock b_bottom(block_width, block_height, tb.texture_w / 3, 0);
	EfficientBlock b_middle(block_width, mid_block_height, tb.texture_w / 3, tb.texture_h / 2 - mid_block_height / 2);
	drawlist_static_.push_back(b_top.xywh());
	drawlist_static_.push_back(b_bottom.xywh());
	drawlist_static_.push_back(b_middle.xywh());
}

/*Lattice targets frequency of 5Hz*/
void SlimBlockchainHandler::queue_lattice_(Toolbox& tb) {
	int block_width = 10;
	int block_height = 10;
	int block_spacing = 30;
	int lattice_width = tb.texture_w / 3;
	int lattice_minx = tb.texture_w / 3 - lattice_width / 2;
	for (float y = 0.f; y + block_height + block_spacing < tb.texture_h; y += block_spacing + block_height) {
		for (float x = lattice_minx; x < lattice_minx + lattice_width; x += block_spacing + block_height) {
			EfficientBlock new_block(block_width, block_height, x, y);
			drawlist_static_.push_back(new_block.xywh());
		}
	}
}

/*Lattice targets frequency of 5Hz*/
void SlimBlockchainHandler::queue_waveguide_(Toolbox& tb) {
	int block_width = 10;
	int block_height = 10;
	int block_spacing = 30;
	int lattice_width = tb.texture_w / 3;
	int lattice_minx = tb.texture_w / 3 - lattice_width / 2;
	for (float y = 0.f; y + block_height + block_spacing < tb.texture_h; y += block_spacing + block_height) {
		if (y < tb.texture_h / 2 && y + block_spacing + block_height > tb.texture_h / 2) {
			continue;
		}
		else if (y > tb.texture_h / 2 && y - block_spacing - block_height < tb.texture_h / 2) {
			continue;
		}
		for (float x = lattice_minx; x < lattice_minx + lattice_width; x += block_spacing + block_height) {
			EfficientBlock new_block(block_width, block_height, x, y);
			drawlist_static_.push_back(new_block.xywh());
		}
	}
}

void SlimBlockchainHandler::queue_ssh_(Toolbox& tb) {
	queued_texture_ = texture_ssh_;
}

void SlimBlockchainHandler::queue_fresnel_(Toolbox& tb) {
	queued_texture_ = texture_ssh_;
}

void SlimBlockchainHandler::xywhs_to_vertices_(const Toolbox& tb, const std::vector<bool>& duplicates,  const std::vector<glm::ivec4>& xywhs, std::vector<float>& vertices) const {
	float texwidth = tb.texture_w;
	float texheight = tb.texture_h;
	vertices.clear();
	vertices.reserve(24 * xywhs.size());
	for (size_t i = 0; i < xywhs.size(); ++i) {
		const glm::ivec4& xywh = xywhs[i];
		if (duplicates[i]) {
			continue;
		}
		float x = xywh.r >= 0 ? xywh.r : 0;
		float y = xywh.g;
		float w = xywh.r >= 0 ? xywh.b : xywh.b + xywh.r;
		float h = xywh.a;
		/*Lower Left Corner*/
		vertices.push_back(2. * (x) / texwidth - 1.f); //x
		vertices.push_back(2. * (y) / texheight - 1.f); //y
		vertices.push_back((x) / texwidth); //u
		vertices.push_back((y) / texheight); //v
		/*Top Right Corner*/
		vertices.push_back(2. * (x + w) / texwidth - 1.f); //x
		vertices.push_back(2. * (y + h) / texheight - 1.f); //y
		vertices.push_back((x + w) / texwidth); //u
		vertices.push_back((y + h) / texheight); //v
		/*Bottom Right Corner*/
		vertices.push_back(2. * (x + w) / texwidth - 1.f); //x
		vertices.push_back(2. * (y) / texheight - 1.f); //y
		vertices.push_back((x + w) / texwidth); //u
		vertices.push_back((y) / texheight); //v
		/*Lower Left Corner*/
		vertices.push_back(2. * (x) / texwidth - 1.f); //x
		vertices.push_back(2. * (y) / texheight - 1.f); //y
		vertices.push_back((x) / texwidth); //u
		vertices.push_back((y) / texheight); //v
		/*Top Right Corner*/
		vertices.push_back(2. * (x + w) / texwidth - 1.f); //x
		vertices.push_back(2. * (y + h) / texheight - 1.f); //y
		vertices.push_back((x + w) / texwidth); //u
		vertices.push_back((y + h) / texheight); //v
		/*Top Left Corner*/
		vertices.push_back(2. * (x) / texwidth - 1.f); //x
		vertices.push_back(2. * (y + h) / texheight - 1.f); //y
		vertices.push_back((x) / texwidth); //u
		vertices.push_back((y + h) / texheight); //v
	}
}

void SlimBlockchainHandler::find_dynamic_duplicates_() {
	drawlist_dynamic_duplicates_ = std::vector<bool>(drawlist_dynamic_.size(), false);
	eraselist_dynamic_duplicates_ = std::vector<bool>(eraselist_dynamic_.size(), false);
	if (drawlist_dynamic_.size() == 0 || eraselist_dynamic_.size() == 0) {
		return;
	}
	for (size_t i_draw = 0; i_draw < drawlist_dynamic_.size(); ++i_draw) {
		const glm::ivec4 dr = drawlist_dynamic_[i_draw];
		for (size_t i_erase = 0; i_erase < eraselist_dynamic_.size(); ++i_erase) {
			const glm::ivec4 er = eraselist_dynamic_[i_erase];
			if (dr.r == er.r && dr.g == er.g && dr.b == er.b && er.a == dr.a) {
				drawlist_dynamic_duplicates_[i_draw] = true;
				//eraselist_dynamic_duplicates_[i_erase] = true;
			}
		}
	}
}

void SlimBlockchainHandler::apply_damping_texture_(GLuint texture) {

}

bool SlimBlockchainHandler::load_damping_texture_(const Toolbox& tb, const std::string filename, GLuint* texture_target) {
	/*Load info about the texture from configuration file*/
	int screen_width, screen_height, texture_width, texture_height;
	int texoffset_left, texoffset_right, texoffset_bottom, texoffset_top;
	std::string config_filename = filename + ".conf";
	std::fstream cfg_file;
	cfg_file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
	try {
		cfg_file.open(config_filename);
		cfg_file >> screen_width;
		cfg_file >> screen_height;
		cfg_file >> texture_width;
		cfg_file >> texture_height;
		cfg_file >> texoffset_left;
		cfg_file >> texoffset_right;
		cfg_file >> texoffset_bottom;
		cfg_file >> texoffset_top;
		cfg_file.close();
	}
	catch (std::ifstream::failure e) {
		std::cout << "Failed to open file " + config_filename << std::endl;
		return false;
	}
	std::cout << "Loaded configuration from " + config_filename;
	std::cout << "\n\tScreen Resolution: " << screen_width << "x" << screen_height;
	std::cout << "\n\tTexture Resolution: " << texture_width << "x" << texture_height;
	std::cout << "\n\tOffset Left: " << texoffset_left;
	std::cout << "\n\tOffset Right: " << texoffset_right;
	std::cout << "\n\tOffset Bottom: " << texoffset_bottom;
	std::cout << "\n\tOffset Top: " << texoffset_top << std::endl;

	/*Check if configuration is compatible with toolbox settings*/
	if (tb.screen_h != screen_height || tb.screen_w != screen_width) {
		std::cout << "Incompatible Toolbox Screen Resolution: " << tb.screen_w << "x" << tb.screen_h << std::endl;
		return false;
	}
	else if (tb.texture_h != texture_height || tb.texture_w != texture_width) {
		std::cout << "Incompatible Toolbox Texture Resolution: " << tb.texture_w << "x" << tb.texture_h << std::endl;
		return false;
	}
	else if (tb.texoffset_left != texoffset_left || tb.texoffset_right != texoffset_right || tb.texoffset_bottom != texoffset_bottom || tb.texoffset_top != texoffset_top) {
		std::cout << "Incompatible Offsets" << std::endl;
		return false;
	}

	/*Load texture data*/
	else {
		std::cout << "Texture is compatible" << std::endl;
	}

	std::cout << "Loading Texture" << std::endl;

	std::fstream dmp_file;
	std::string damping_filename = filename + ".texture";
	std::vector<float> data_target;
	data_target.reserve(4 * texture_width * texture_height);
	dmp_file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
	try {
		dmp_file.open(damping_filename);

		float value;
		for (size_t i = 0; i < 4 * texture_width * texture_height; ++i) {
			/*If the file is finished at this point, we're fucked*/
			if (dmp_file.peek() == EOF) {
				std::cout << "Unexpected End Of Damping File Encountered" << std::endl;
				data_target.clear();
				return false;
			}
			dmp_file >> value;
			data_target.push_back(value);
		}

		dmp_file.close();
	}
	catch (std::ifstream::failure e) {
		std::cout << "Failed to open file " + damping_filename << std::endl;
		return false;
	}

	/*Generate and upload texture*/
	glGenTextures(1, texture_target);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, *texture_target);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, texture_width, texture_height, 0, GL_RGBA, GL_FLOAT, data_target.data());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_2D, 0);


	return true;
}